@@ -18,6 +18,7 @@
 {-# LANGUAGE StandaloneDeriving #-}
 {-# LANGUAGE TupleSections #-}
 {-# LANGUAGE TypeFamilies #-}
+{-# LANGUAGE TypeOperators #-}
 {-# LANGUAGE UndecidableInstances #-} -- MonadBase
 #if __GLASGOW_HASKELL__ >= 806
 {-# LANGUAGE DerivingVia #-}
@@ -129,6 +130,7 @@ module Hedgehog.Internal.Gen (
 
   -- ** Combinations & Permutations
   , subsequence
+  , subset
   , shuffle
   , shuffleSeq
 
@@ -215,6 +217,7 @@ import qualified Data.Semigroup as Semigroup
 import           Data.Sequence (Seq)
 import qualified Data.Sequence as Seq
 import           Data.Set (Set)
+import qualified Data.Set as Set
 import           Data.Text (Text)
 import qualified Data.Text as Text
 import qualified Data.Text.Encoding as Text
@@ -471,9 +474,15 @@ instance (Monad m, Semigroup a) => Semigroup (GenT m a) where
   (<>) =
     liftA2 (Semigroup.<>)
 
-instance (Monad m, Monoid a) => Monoid (GenT m a) where
-  mappend =
-    liftA2 mappend
+instance (
+  Monad m, Monoid a
+#if !MIN_VERSION_base(4,11,0)
+  , Semigroup a
+#endif
+         ) => Monoid (GenT m a) where
+#if !MIN_VERSION_base(4,11,0)
+  mappend = (Semigroup.<>)
+#endif
 
   mempty =
     return mempty
@@ -1655,10 +1664,37 @@ subterm3 gx gy gz f =
 
 -- | Generates a random subsequence of a list.
 --
+-- For example:
+--
+-- @
+-- Gen.print (Gen.subsequence [1..5])
+-- @
+--
+--   > === Outcome ===
+--   > [1,2,4]
+--   > === Shrinks ===
+--   > []
+--   > [2,4]
+--   > [1,4]
+--   > [1,2]
+--
 subsequence :: MonadGen m => [a] -> m [a]
 subsequence xs =
   shrink Shrink.list $ filterM (const bool_) xs
 
+-- | Generates a random subset of a set.
+--
+--  /This shrinks towards the empty set./
+--
+subset :: MonadGen m => Set a -> m (Set a)
+-- Set.fromDistinctAscList has an unchecked precondition that the list
+-- must be strictly ascending. This precondition is satisfied because
+-- Set.toAscList produces a strictly ascending list, and the 'subsequence'
+-- generator only removes elements from the list; it never adds or
+-- rearranges elements, so the strictly ascending property is undisturbed.
+subset =
+  fmap Set.fromDistinctAscList . subsequence . Set.toAscList
+
 -- | Generates a random permutation of a list.
 --
 --   /This shrinks towards the order of the list being identical to the input/
