@@ -34,11 +34,14 @@ module Hedgehog.Internal.Property (
   , DiscardCount(..)
   , ShrinkLimit(..)
   , ShrinkCount(..)
+  , Skip(..)
+  , ShrinkPath(..)
   , ShrinkRetries(..)
   , withTests
   , withDiscards
   , withShrinks
   , withRetries
+  , withSkip
   , property
   , test
   , forAll
@@ -47,6 +50,10 @@ module Hedgehog.Internal.Property (
   , forAllWithT
   , defaultMinTests
   , discard
+  , skipCompress
+  , shrinkPathCompress
+  , skipDecompress
+  , shrinkPathDecompress
 
   -- * Group
   , Group(..)
@@ -165,7 +172,7 @@ import           Data.Map (Map)
 import qualified Data.Map.Strict as Map
 import           Data.Number.Erf (invnormcdf)
 import qualified Data.List as List
-import           Data.String (IsString)
+import           Data.String (IsString(..))
 import           Data.Ratio ((%))
 import           Data.Typeable (typeOf)
 
@@ -179,6 +186,9 @@ import           Hedgehog.Internal.Source
 
 import           Language.Haskell.TH.Syntax (Lift)
 
+import qualified Numeric
+
+import           Text.Read (readMaybe)
 
 ------------------------------------------------------------------------
 
@@ -273,6 +283,10 @@ data PropertyConfig =
     , propertyShrinkLimit :: !ShrinkLimit
     , propertyShrinkRetries :: !ShrinkRetries
     , propertyTerminationCriteria :: !TerminationCriteria
+
+    -- | If this is 'Nothing', we take the Skip from the environment variable
+    --   @HEDGEHOG_SKIP@.
+    , propertySkip :: Maybe Skip
     } deriving (Eq, Ord, Show, Lift)
 
 -- | The number of successful tests that need to be run before a property test
@@ -292,7 +306,7 @@ newtype TestLimit =
 --
 newtype TestCount =
   TestCount Int
-  deriving (Eq, Ord, Show, Num, Enum, Real, Integral)
+  deriving (Eq, Ord, Show, Num, Enum, Real, Integral, Lift)
 
 -- | The number of tests a property had to discard.
 --
@@ -331,6 +345,169 @@ newtype ShrinkCount =
   ShrinkCount Int
   deriving (Eq, Ord, Show, Num, Enum, Real, Integral)
 
+-- | Where to start running a property's tests.
+--
+data Skip =
+  -- | Don't skip anything.
+  --
+    SkipNothing
+
+  -- | Skip to a specific test number. If it fails, shrink as normal. If it
+  --   passes, move on to the next test. Coverage checks are disabled.
+  --
+  | SkipToTest TestCount
+
+  -- | Skip to a specific test number and shrink state. If it fails, stop
+  --   without shrinking further. If it passes, the property will pass without
+  --   running any more tests.
+  --
+  --   Due to implementation details, all intermediate shrink states - those on
+  --   the direct path from the original test input to the target state - will
+  --   be tested too, and their results discarded.
+  --
+  | SkipToShrink TestCount ShrinkPath
+  deriving (Eq, Ord, Show, Lift)
+
+-- | We use this instance to support usage like
+--
+-- @
+--   withSkip "3:aB"
+-- @
+--
+--   It throws an error if the input is not a valid compressed 'Skip'.
+--
+instance IsString Skip where
+  fromString s =
+    case skipDecompress s of
+      Nothing ->
+        error $ "fromString: Not a valid Skip: " ++ s
+      Just skip ->
+        skip
+
+-- | The path taken to reach a shrink state.
+--
+newtype ShrinkPath =
+  ShrinkPath [Int]
+  deriving (Eq, Ord, Show, Lift)
+
+-- | Compress a Skip into a hopefully-short alphanumeric string.
+--
+--   The bit that might be long is the 'ShrinkPath' in 'SkipToShrink'. For that,
+--   we encode the path components in base 26, alternating between uppercase and
+--   lowercase alphabets to distinguish list elements. Additionally when we have
+--   runs of equal components, we use the normal base 10 encoding to indicate
+--   the length.
+--
+--   This gives something which is hopefully quite short, but a human can
+--   roughly interpret it by eyeball.
+--
+skipCompress :: Skip -> String
+skipCompress = \case
+  SkipNothing ->
+    ""
+  SkipToTest (TestCount n) ->
+    show n
+  SkipToShrink (TestCount n) sp ->
+    show n ++ ":" ++ shrinkPathCompress sp
+
+-- | Compress a 'ShrinkPath' into a hopefully-short alphanumeric string.
+--
+--   We encode the path components in base 26, alternating between uppercase and
+--   lowercase alphabets to distinguish list elements. Additionally when we have
+--   runs of equal components, we use the normal base 10 encoding to indicate
+--   the length.
+shrinkPathCompress :: ShrinkPath -> String
+shrinkPathCompress (ShrinkPath sp) =
+  let
+    groups = List.map (\l -> (head l, length l)) $ List.group sp
+  in
+    (mconcat
+      $ zipWith
+          (\alphabet (loc, count) ->
+              Numeric.showIntAtBase 26 (alphabet !!) loc
+              <> if count == 1 then mempty else shows count
+          )
+          (cycle [['a'..'z'], ['A'..'Z']])
+          groups
+    )
+      ""
+
+-- | Decompress a 'Skip'.
+--
+--   This satisfies
+--
+-- @
+--   skipDecompress (skipCompress a) == Just a
+-- @
+--
+skipDecompress :: String -> Maybe Skip
+skipDecompress str =
+  if null str then
+    Just SkipNothing
+  else do
+    let
+      (tcStr, spStr)
+        = span (/= ':') str
+    tc <- TestCount <$> readMaybe tcStr
+    if null spStr then
+      Just $ SkipToTest tc
+    else do
+      sp <- shrinkPathDecompress $ drop 1 spStr
+      Just $ SkipToShrink tc sp
+
+-- | Decompress a 'ShrinkPath'.
+--
+--   This satisfies
+--
+-- @
+--   shrinkPathDecompress (shrinkPathCompress a) == Just a
+-- @
+shrinkPathDecompress :: String -> Maybe ShrinkPath
+shrinkPathDecompress str =
+  let
+    isDigit c = '0' <= c && c <= '9'
+    isLower c = 'a' <= c && c <= 'z'
+    isUpper c = 'A' <= c && c <= 'Z'
+    classifyChar c = (isDigit c, isLower c, isUpper c)
+
+    readSNum "" = []
+    readSNum s@(c1:_) =
+      if isDigit c1 then
+        Numeric.readInt 10 isDigit (\c -> fromEnum c - fromEnum '0') s
+      else if isLower c1 then
+        Numeric.readInt 26 isLower (\c -> fromEnum c - fromEnum 'a') s
+      else if isUpper c1 then
+        Numeric.readInt 26 isUpper (\c -> fromEnum c - fromEnum 'A') s
+      else
+        []
+
+    readNumMaybe s =
+      case readSNum s of
+        [(num, "")] -> Just num
+        _ -> Nothing
+
+    spGroups :: [(Maybe Int, Maybe Int)] =
+      let
+        go [] =
+          []
+        go (c1:cs) =
+          let
+            (hd, tl1) =
+              span (\c -> classifyChar c == classifyChar c1) cs
+            (digs, tl2) =
+              span isDigit tl1
+          in
+            ( readNumMaybe (c1:hd)
+            , readNumMaybe $ if null digs then "1" else digs
+            )
+            : go tl2
+      in
+        go str
+  in do
+    sp <- concat <$>
+      traverse (\(mNum, mCount) -> replicate <$> mCount <*> mNum) spGroups
+    Just $ ShrinkPath sp
+
 -- | The number of times to re-run a test during shrinking. This is useful if
 --   you are testing something which fails non-deterministically and you want to
 --   increase the change of getting a good shrink.
@@ -991,6 +1168,8 @@ defaultConfig =
         0
     , propertyTerminationCriteria =
         NoConfidenceTermination defaultMinTests
+    , propertySkip =
+        Nothing
     }
 
 -- | The minimum amount of tests to run for a 'Property'
@@ -1077,6 +1256,12 @@ withRetries :: ShrinkRetries -> Property -> Property
 withRetries n =
   mapConfig $ \config -> config { propertyShrinkRetries = n }
 
+-- | Set the target that a property will skip to before it starts to run.
+--
+withSkip :: Skip -> Property -> Property
+withSkip s =
+  mapConfig $ \config -> config { propertySkip = Just s }
+
 -- | Creates a property with the default configuration.
 --
 property :: HasCallStack => PropertyT IO () -> Property
@@ -1248,8 +1433,8 @@ journalCoverage (Journal logs) =
 --    prop_with_coverage =
 --      property $ do
 --        match <- forAll Gen.bool
---        cover 30 "True" $ match
---        cover 30 "False" $ not match
+--        cover 30 \"True\" $ match
+--        cover 30 \"False\" $ not match
 -- @
 --
 --   The example above requires a minimum of 30% coverage for both
