@@ -46,6 +46,9 @@ import qualified Data.List as List
 import           Data.Map (Map)
 import qualified Data.Map as Map
 import           Data.Maybe (mapMaybe, catMaybes)
+#if !MIN_VERSION_base(4,11,0)
+import qualified Data.Semigroup as Semigroup
+#endif
 import           Data.Traversable (for)
 
 import           Hedgehog.Internal.Config
@@ -59,11 +62,10 @@ import           Hedgehog.Internal.Property (ShrinkCount(..), PropertyCount(..))
 import           Hedgehog.Internal.Property (TestCount(..), DiscardCount(..))
 import           Hedgehog.Internal.Property (coverPercentage, coverageFailures)
 import           Hedgehog.Internal.Property (labelCovered)
+import           Hedgehog.Internal.Property (ShrinkPath(..), skipCompress)
 
-import           Hedgehog.Internal.Seed (Seed)
 import           Hedgehog.Internal.Show
 import           Hedgehog.Internal.Source
-import           Hedgehog.Range (Size)
 
 import           System.Console.ANSI (ColorIntensity(..), Color(..))
 import           System.Console.ANSI (ConsoleLayer(..), ConsoleIntensity(..))
@@ -89,9 +91,8 @@ data FailedAnnotation =
 
 data FailureReport =
   FailureReport {
-      failureSize :: !Size
-    , failureSeed :: !Seed
-    , failureShrinks :: !ShrinkCount
+      failureShrinks :: !ShrinkCount
+    , failureShrinkPath :: !ShrinkPath
     , failureCoverage :: !(Maybe (Coverage CoverCount))
     , failureAnnotations :: ![FailedAnnotation]
     , failureLocation :: !(Maybe Span)
@@ -125,6 +126,7 @@ data Report a =
       reportTests :: !TestCount
     , reportDiscards :: !DiscardCount
     , reportCoverage :: !(Coverage CoverCount)
+    , reportSeed :: !Seed
     , reportStatus :: !a
     } deriving (Show, Functor, Foldable, Traversable)
 
@@ -140,9 +142,14 @@ data Summary =
     } deriving (Show)
 
 instance Monoid Summary where
+#if !MIN_VERSION_base(4,11,0)
+  mappend = (Semigroup.<>)
+#endif
   mempty =
     Summary 0 0 0 0 0
-  mappend (Summary x1 x2 x3 x4 x5) (Summary y1 y2 y3 y4 y5) =
+
+instance Semigroup Summary where
+  Summary x1 x2 x3 x4 x5 <> Summary y1 y2 y3 y4 y5 =
     Summary
       (x1 + y1)
       (x2 + y2)
@@ -150,9 +157,6 @@ instance Monoid Summary where
       (x4 + y4)
       (x5 + y5)
 
-instance Semigroup Summary where
-  (<>) = mappend
-
 -- | Construct a summary from a single result.
 --
 fromResult :: Result -> Summary
@@ -263,16 +267,15 @@ takeFootnote = \case
     Nothing
 
 mkFailure ::
-     Size
-  -> Seed
-  -> ShrinkCount
+     ShrinkCount
+  -> ShrinkPath
   -> Maybe (Coverage CoverCount)
   -> Maybe Span
   -> String
   -> Maybe Diff
   -> [Log]
   -> FailureReport
-mkFailure size seed shrinks mcoverage location message diff logs =
+mkFailure shrinks shrinkPath mcoverage location message diff logs =
   let
     inputs =
       mapMaybe takeAnnotation logs
@@ -280,7 +283,7 @@ mkFailure size seed shrinks mcoverage location message diff logs =
     footnotes =
       mapMaybe takeFootnote logs
   in
-    FailureReport size seed shrinks mcoverage inputs location message diff footnotes
+    FailureReport shrinks shrinkPath mcoverage inputs location message diff footnotes
 
 ------------------------------------------------------------------------
 -- Pretty Printing
@@ -322,6 +325,18 @@ ppShrinkCount = \case
   ShrinkCount n ->
     ppShow n <+> "shrinks"
 
+-- | Render a compressed 'Skip'.
+--
+ppSkip :: Skip -> Doc a
+ppSkip =
+  WL.text . skipCompress
+
+-- | Render a compressed 'Skip', such that it can be read back in.
+--
+ppSkipReadable :: Skip -> Doc a
+ppSkipReadable =
+  WL.text . show . skipCompress
+
 ppRawPropertyCount :: PropertyCount -> Doc a
 ppRawPropertyCount (PropertyCount n) =
   ppShow n
@@ -576,15 +591,15 @@ ppDeclaration decl =
       in
         WL.vsep (ppLocation : ppLines)
 
-ppReproduce :: Maybe PropertyName -> Size -> Seed -> Doc Markup
-ppReproduce name size seed =
+ppReproduce :: Maybe PropertyName -> Seed -> Skip -> Doc Markup
+ppReproduce name seed skip =
   WL.vsep [
       markup ReproduceHeader
         "This failure can be reproduced by running:"
     , gutter ReproduceGutter . markup ReproduceSource $
-        "recheck" <+>
-        WL.text (showsPrec 11 size "") <+>
+        "recheckAt" <+>
         WL.text (showsPrec 11 seed "") <+>
+        ppSkipReadable skip <+>
         maybe "<property>" (WL.text . unPropertyName) name
     ]
 
@@ -607,8 +622,8 @@ ppTextLines :: String -> [Doc Markup]
 ppTextLines =
   fmap WL.text . List.lines
 
-ppFailureReport :: MonadIO m => Maybe PropertyName -> TestCount -> FailureReport -> m [Doc Markup]
-ppFailureReport name tests (FailureReport size seed _ mcoverage inputs0 mlocation0 msg mdiff msgs0) = do
+ppFailureReport :: MonadIO m => Maybe PropertyName -> TestCount -> Seed -> FailureReport -> m [Doc Markup]
+ppFailureReport name tests seed (FailureReport _ shrinkPath mcoverage inputs0 mlocation0 msg mdiff msgs0) = do
   let
     basic =
       -- Move the failure message to the end section if we have
@@ -680,7 +695,10 @@ ppFailureReport name tests (FailureReport size seed _ mcoverage inputs0 mlocatio
         f xs
 
     bottom =
-      maybe [ppReproduce name size seed] (const []) mcoverage
+      maybe
+        [ppReproduce name seed (SkipToShrink tests shrinkPath)]
+        (const [])
+        mcoverage
 
   pure .
     whenSome (mempty :) .
@@ -708,7 +726,7 @@ ppName = \case
     WL.text name
 
 ppProgress :: MonadIO m => Maybe PropertyName -> Report Progress -> m (Doc Markup)
-ppProgress name (Report tests discards coverage status) =
+ppProgress name (Report tests discards coverage _ status) =
   case status of
     Running ->
       pure . WL.vsep $ [
@@ -731,10 +749,10 @@ ppProgress name (Report tests discards coverage status) =
         "(shrinking)"
 
 ppResult :: MonadIO m => Maybe PropertyName -> Report Result -> m (Doc Markup)
-ppResult name (Report tests discards coverage result) = do
+ppResult name (Report tests discards coverage seed result) = do
   case result of
     Failed failure -> do
-      pfailure <- ppFailureReport name tests failure
+      pfailure <- ppFailureReport name tests seed failure
       pure . WL.vsep $ [
           icon FailedIcon '✗' . WL.align . WL.annotate FailedText $
             ppName name <+>
@@ -742,7 +760,9 @@ ppResult name (Report tests discards coverage result) = do
             "after" <+>
             ppTestCount tests <>
             ppShrinkDiscard (failureShrinks failure) discards <>
-            "."
+            "." <#>
+            "shrink path:" <+>
+            ppSkip (SkipToShrink tests $ failureShrinkPath failure)
         ] ++
         ppCoverage tests coverage ++
         pfailure
