@@ -13,6 +13,7 @@ module Hedgehog.Internal.Runner (
   -- * Running Individual Properties
     check
   , recheck
+  , recheckAt
 
   -- * Running Groups of Properties
   , RunnerConfig(..)
@@ -30,6 +31,7 @@ import           Control.Concurrent.STM (TVar, atomically)
 import qualified Control.Concurrent.STM.TVar as TVar
 import           Control.Monad.Catch (MonadCatch(..), catchAll)
 import           Control.Monad.IO.Class (MonadIO(..))
+import           Data.Maybe (isJust)
 
 import           Hedgehog.Internal.Config
 import           Hedgehog.Internal.Gen (evalGenT)
@@ -39,16 +41,16 @@ import           Hedgehog.Internal.Property (Group(..), GroupName(..))
 import           Hedgehog.Internal.Property (Journal(..), Coverage(..), CoverCount(..))
 import           Hedgehog.Internal.Property (Property(..), PropertyConfig(..), PropertyName(..))
 import           Hedgehog.Internal.Property (PropertyT(..), Failure(..), runTestT)
-import           Hedgehog.Internal.Property (ShrinkLimit, ShrinkRetries, withTests)
+import           Hedgehog.Internal.Property (ShrinkLimit, ShrinkRetries, withTests, withSkip)
 import           Hedgehog.Internal.Property (TerminationCriteria(..))
 import           Hedgehog.Internal.Property (TestCount(..), PropertyCount(..))
 import           Hedgehog.Internal.Property (confidenceSuccess, confidenceFailure)
 import           Hedgehog.Internal.Property (coverageSuccess, journalCoverage)
 import           Hedgehog.Internal.Property (defaultMinTests)
+import           Hedgehog.Internal.Property (ShrinkPath(..))
 import           Hedgehog.Internal.Queue
 import           Hedgehog.Internal.Region
 import           Hedgehog.Internal.Report
-import           Hedgehog.Internal.Seed (Seed)
 import qualified Hedgehog.Internal.Seed as Seed
 import           Hedgehog.Internal.Tree (TreeT(..), NodeT(..))
 import           Hedgehog.Range (Size)
@@ -71,6 +73,9 @@ data RunnerConfig =
       --   the environment.
     , runnerColor :: !(Maybe UseColor)
 
+      -- | The seed to use. 'Nothing' means detect from the environment.
+    , runnerSeed :: !(Maybe Seed)
+
       -- | How verbose to be in the runner output. 'Nothing' means detect from
       --   the environment.
     , runnerVerbosity :: !(Maybe Verbosity)
@@ -114,40 +119,88 @@ runTreeN n m = do
 
 takeSmallest ::
      MonadIO m
-  => Size
-  -> Seed
-  -> ShrinkCount
+  => ShrinkCount
+  -> ShrinkPath
   -> ShrinkLimit
   -> ShrinkRetries
   -> (Progress -> m ())
   -> NodeT m (Maybe (Either Failure (), Journal))
   -> m Result
-takeSmallest size seed shrinks slimit retries updateUI = \case
-  NodeT Nothing _ ->
-    pure GaveUp
+takeSmallest shrinks0 (ShrinkPath shrinkPath0) slimit retries updateUI =
+  let
+    loop shrinks revShrinkPath = \case
+      NodeT Nothing _ ->
+        pure GaveUp
 
-  NodeT (Just (x, (Journal logs))) xs ->
-    case x of
-      Left (Failure loc err mdiff) -> do
-        let
-          failure =
-            mkFailure size seed shrinks Nothing loc err mdiff (reverse logs)
+      NodeT (Just (x, (Journal logs))) xs ->
+        case x of
+          Left (Failure loc err mdiff) -> do
+            let
+              shrinkPath =
+                ShrinkPath $ reverse revShrinkPath
+              failure =
+                mkFailure shrinks shrinkPath Nothing loc err mdiff (reverse logs)
 
-        updateUI $ Shrinking failure
+            updateUI $ Shrinking failure
 
-        if shrinks >= fromIntegral slimit then
-          -- if we've hit the shrink limit, don't shrink any further
-          pure $ Failed failure
-        else
-          findM xs (Failed failure) $ \m -> do
-            o <- runTreeN retries m
-            if isFailure o then
-              Just <$> takeSmallest size seed (shrinks + 1) slimit retries updateUI o
+            if shrinks >= fromIntegral slimit then
+              -- if we've hit the shrink limit, don't shrink any further
+              pure $ Failed failure
             else
-              return Nothing
+              findM (zip [0..] xs) (Failed failure) $ \(n, m) -> do
+                o <- runTreeN retries m
+                if isFailure o then
+                  Just <$> loop (shrinks + 1) (n : revShrinkPath) o
+                else
+                  return Nothing
 
-      Right () ->
-        return OK
+          Right () ->
+            return OK
+  in
+    loop shrinks0 (reverse shrinkPath0)
+
+-- | Follow a given shrink path, instead of searching exhaustively. Assume that
+-- the end of the path is minimal, and don't try to shrink any further than
+-- that.
+--
+-- This evaluates the test for all the shrinks on the path, but not ones
+-- off-path. Because the generator is mixed with the test code, it's probably
+-- not possible to avoid this.
+skipToShrink ::
+     MonadIO m
+  => ShrinkPath
+  -> (Progress -> m ())
+  -> NodeT m (Maybe (Either Failure (), Journal))
+  -> m Result
+skipToShrink (ShrinkPath shrinkPath) updateUI =
+  let
+    loop shrinks [] = \case
+      NodeT Nothing _ ->
+        pure GaveUp
+
+      NodeT (Just (x, (Journal logs))) _ ->
+        case x of
+          Left (Failure loc err mdiff) -> do
+            let
+              failure =
+                mkFailure shrinks (ShrinkPath shrinkPath) Nothing loc err mdiff (reverse logs)
+
+            updateUI $ Shrinking failure
+            pure $ Failed failure
+
+          Right () ->
+            return OK
+
+    loop shrinks (s0:ss) = \case
+      NodeT _ xs ->
+        case drop s0 xs of
+          [] ->
+            pure GaveUp
+          (x:_) -> do
+            o <- runTreeT x
+            loop (shrinks + 1) ss o
+  in
+    loop 0 shrinkPath
 
 checkReport ::
      forall m.
@@ -159,8 +212,19 @@ checkReport ::
   -> PropertyT m ()
   -> (Report Progress -> m ())
   -> m (Report Result)
-checkReport cfg size0 seed0 test0 updateUI =
+checkReport cfg size0 seed0 test0 updateUI = do
+  skip <- liftIO $ resolveSkip $ propertySkip cfg
+
   let
+    (mSkipToTest, mSkipToShrink) =
+      case skip of
+        SkipNothing ->
+          (Nothing, Nothing)
+        SkipToTest t ->
+          (Just t, Nothing)
+        SkipToShrink t s ->
+          (Just t, Just s)
+
     test =
       catchAll test0 (fail . show)
 
@@ -196,7 +260,7 @@ checkReport cfg size0 seed0 test0 updateUI =
       -> Coverage CoverCount
       -> m (Report Result)
     loop !tests !discards !size !seed !coverage0 = do
-      updateUI $ Report tests discards coverage0 Running
+      updateUI $ Report tests discards coverage0 seed0 Running
 
       let
         coverageReached =
@@ -219,13 +283,12 @@ checkReport cfg size0 seed0 test0 updateUI =
           coverageSuccess tests coverage0
 
         successReport =
-          Report tests discards coverage0 OK
+          Report tests discards coverage0 seed0 OK
 
         failureReport message =
-          Report tests discards coverage0 . Failed $ mkFailure
-            size
-            seed
+          Report tests discards coverage0 seed0 . Failed $ mkFailure
             0
+            (ShrinkPath [])
             (Just coverage0)
             Nothing
             message
@@ -247,53 +310,69 @@ checkReport cfg size0 seed0 test0 updateUI =
         -- at this point, we know that enough tests have been run in order to
         -- make a decision on if this was a successful run or not
         --
-        -- If we have early termination, then we need to check coverageReached / coverageUnreachable
-        pure $ case terminationCriteria of
-          EarlyTermination _ _ -> confidenceReport
-          NoEarlyTermination _ _ -> confidenceReport
-          NoConfidenceTermination _ ->
-            if labelsCovered then
-              successReport
-            else
-              failureReport $
-                "Labels not sufficently covered after " <> show tests <> " tests"
+        -- If we have early termination, then we need to check coverageReached /
+        -- coverageUnreachable. If we skip tests, we ignore coverage.
+        if isJust mSkipToTest then
+          pure successReport
+        else
+          pure $ case terminationCriteria of
+            EarlyTermination _ _ -> confidenceReport
+            NoEarlyTermination _ _ -> confidenceReport
+            NoConfidenceTermination _ ->
+              if labelsCovered then
+                successReport
+              else
+                failureReport $
+                  "Labels not sufficently covered after " <> show tests <> " tests"
 
       else if discards >= fromIntegral (propertyDiscardLimit cfg) then
         -- we've hit the discard limit, give up
-        pure $ Report tests discards coverage0 GaveUp
+        pure $ Report tests discards coverage0 seed0 GaveUp
 
       else
         case Seed.split seed of
-          (s0, s1) -> do
-            node@(NodeT x _) <-
-              runTreeT . evalGenT size s0 . runTestT $ unPropertyT test
-            case x of
-              Nothing ->
-                loop tests (discards + 1) (size + 1) s1 coverage0
+          (s0, s1) -> case (mSkipToTest, mSkipToShrink) of
+            -- If the report says failed "after 32 tests", the test number that
+            -- failed was 31, but we want the user to be able to skip to 32 and
+            -- start with the one that failed.
+            (Just n, _) | n > tests + 1 ->
+              loop (tests + 1) discards (size + 1) s1 coverage0
+            (Just _, Just shrinkPath) -> do
+              node <-
+                runTreeT . evalGenT size s0 . runTestT $ unPropertyT test
+              let
+                mkReport =
+                  Report (tests + 1) discards coverage0 seed0
+              mkReport <$> skipToShrink shrinkPath (updateUI . mkReport) node
+            _ -> do
+              node@(NodeT x _) <-
+                runTreeT . evalGenT size s0 . runTestT $ unPropertyT test
+              case x of
+                Nothing ->
+                  loop tests (discards + 1) (size + 1) s1 coverage0
 
-              Just (Left _, _) ->
-                let
-                  mkReport =
-                    Report (tests + 1) discards coverage0
-                in
-                  fmap mkReport $
-                    takeSmallest
-                      size
-                      seed
-                      0
-                      (propertyShrinkLimit cfg)
-                      (propertyShrinkRetries cfg)
-                      (updateUI . mkReport)
-                      node
+                Just (Left _, _) ->
+                  let
+                    mkReport =
+                      Report (tests + 1) discards coverage0 seed0
+                  in
+                    fmap mkReport $
+                      takeSmallest
+                        0
+                        (ShrinkPath [])
+                        (propertyShrinkLimit cfg)
+                        (propertyShrinkRetries cfg)
+                        (updateUI . mkReport)
+                        node
 
-              Just (Right (), journal) ->
-                let
-                  coverage =
-                    journalCoverage journal <> coverage0
-                in
-                  loop (tests + 1) discards (size + 1) s1 coverage
-  in
-    loop 0 0 size0 seed0 mempty
+                Just (Right (), journal) ->
+                  let
+                    coverage =
+                      journalCoverage journal <> coverage0
+                  in
+                    loop (tests + 1) discards (size + 1) s1 coverage
+
+  loop 0 0 size0 seed0 mempty
 
 checkRegion ::
      MonadIO m
@@ -331,10 +410,11 @@ checkNamed ::
   => Region
   -> UseColor
   -> Maybe PropertyName
+  -> Maybe Seed
   -> Property
   -> m (Report Result)
-checkNamed region color name prop = do
-  seed <- liftIO Seed.random
+checkNamed region color name mseed prop = do
+  seed <- resolveSeed mseed
   checkRegion region color name 0 seed prop
 
 -- | Check a property.
@@ -343,7 +423,7 @@ check :: MonadIO m => Property -> m Bool
 check prop = do
   color <- detectColor
   liftIO . displayRegion $ \region ->
-    (== OK) . reportStatus <$> checkNamed region color Nothing prop
+    (== OK) . reportStatus <$> checkNamed region color Nothing Nothing prop
 
 -- | Check a property using a specific size and seed.
 --
@@ -355,6 +435,14 @@ recheck size seed prop0 = do
     checkRegion region color Nothing size seed prop
   pure ()
 
+recheckAt :: MonadIO m => Seed -> Skip -> Property -> m ()
+recheckAt seed skip prop0 = do
+  color <- detectColor
+  let prop = withSkip skip prop0
+  _ <- liftIO . displayRegion $ \region ->
+    checkRegion region color Nothing 0 seed prop
+  pure ()
+
 -- | Check a group of properties using the specified runner config.
 --
 checkGroup :: MonadIO m => RunnerConfig -> Group -> m Bool
@@ -373,9 +461,10 @@ checkGroup config (Group group props) =
 
     putStrLn $ "━━━ " ++ unGroupName group ++ " ━━━"
 
+    seed <- resolveSeed (runnerSeed config)
     verbosity <- resolveVerbosity (runnerVerbosity config)
     color <- resolveColor (runnerColor config)
-    summary <- checkGroupWith n verbosity color props
+    summary <- checkGroupWith n verbosity color seed props
 
     pure $
       summaryFailed summary == 0 &&
@@ -390,9 +479,10 @@ checkGroupWith ::
      WorkerCount
   -> Verbosity
   -> UseColor
+  -> Seed
   -> [(PropertyName, Property)]
   -> IO Summary
-checkGroupWith n verbosity color props =
+checkGroupWith n verbosity color seed props =
   displayRegion $ \sregion -> do
     svar <- atomically . TVar.newTVar $ mempty { summaryWaiting = PropertyCount (length props) }
 
@@ -430,7 +520,7 @@ checkGroupWith n verbosity color props =
     summary <-
       fmap (mconcat . fmap (fromResult . reportStatus)) $
         runTasks n props start finish finalize $ \(name, prop, region) -> do
-          result <- checkNamed region color (Just name) prop
+          result <- checkNamed region color (Just name) (Just seed) prop
           updateSummary sregion svar color
             (<> fromResult (reportStatus result))
           pure result
@@ -463,6 +553,8 @@ checkSequential =
           Just 1
       , runnerColor =
           Nothing
+      , runnerSeed =
+          Nothing
       , runnerVerbosity =
           Nothing
       }
@@ -497,6 +589,8 @@ checkParallel =
           Nothing
       , runnerColor =
           Nothing
+      , runnerSeed =
+          Nothing
       , runnerVerbosity =
           Nothing
       }
