@@ -71,10 +71,10 @@ import qualified Data.Map as Map
 import qualified Data.Maybe as Maybe
 import           Data.Typeable (Typeable, TypeRep, Proxy(..), typeRep)
 
+import           Hedgehog.Internal.Barbie (FunctorB(..), TraversableB(..))
 import           Hedgehog.Internal.Distributive (distributeT)
 import           Hedgehog.Internal.Gen (MonadGen, GenT, GenBase)
 import qualified Hedgehog.Internal.Gen as Gen
-import           Hedgehog.Internal.HTraversable (HTraversable(..))
 import           Hedgehog.Internal.Opaque (Opaque(..))
 import           Hedgehog.Internal.Property (MonadTest(..), Test, evalEither, evalM, success, runTest, failWith, annotate)
 import           Hedgehog.Internal.Range (Range)
@@ -202,8 +202,12 @@ instance (Show a, Show1 v) => Show (Var a v) where
       showString "Var " .
       showsPrec1 11 x
 
-instance HTraversable (Var a) where
-  htraverse f (Var v) =
+instance FunctorB (Var a) where
+  bmap f (Var v) =
+    Var (f v)
+
+instance TraversableB (Var a) where
+  btraverse f (Var v) =
     fmap Var (f v)
 
 ------------------------------------------------------------------------
@@ -262,9 +266,9 @@ reifyEnvironment (Environment vars) (Symbolic n) =
 
 -- | Convert a symbolic structure to a concrete one, using the provided environment.
 --
-reify :: HTraversable t => Environment -> t Symbolic -> Either EnvironmentError (t Concrete)
+reify :: TraversableB t => Environment -> t Symbolic -> Either EnvironmentError (t Concrete)
 reify vars =
-  htraverse (reifyEnvironment vars)
+  btraverse (reifyEnvironment vars)
 
 ------------------------------------------------------------------------
 -- Callbacks
@@ -377,7 +381,7 @@ callbackEnsure callbacks s0 s i o =
 --
 data Command gen m (state :: (Type -> Type) -> Type) =
   forall input output.
-  (HTraversable input, Show (input Symbolic), Show output, Typeable output) =>
+  (TraversableB input, Show (input Symbolic), Show output, Typeable output) =>
   Command {
     -- | A generator which provides random arguments for a command. If the
     --   command cannot be executed in the current state, it should return
@@ -409,7 +413,7 @@ commandGenOK (Command inputGen _ _) state =
 --
 data Action m (state :: (Type -> Type) -> Type) =
   forall input output.
-  (HTraversable input, Show (input Symbolic), Show output) =>
+  (TraversableB input, Show (input Symbolic), Show output) =>
   Action {
       actionInput ::
         input Symbolic
@@ -457,19 +461,19 @@ insertSymbolic s =
 -- | Collects all the symbolic values in a data structure and produces a set of
 --   all the variables they refer to.
 --
-takeVariables :: forall t. HTraversable t => t Symbolic -> Map Name TypeRep
+takeVariables :: forall t. TraversableB t => t Symbolic -> Map Name TypeRep
 takeVariables xs =
   let
     go x = do
       modify (insertSymbolic x)
       pure x
   in
-    flip execState Map.empty $ htraverse go xs
+    flip execState Map.empty $ btraverse go xs
 
 -- | Checks that the symbolic values in the data structure refer only to the
 --   variables in the provided set, and that they are of the correct type.
 --
-variablesOK :: HTraversable t => t Symbolic -> Map Name TypeRep -> Bool
+variablesOK :: TraversableB t => t Symbolic -> Map Name TypeRep -> Bool
 variablesOK xs allowed =
   let
     vars =
@@ -544,6 +548,12 @@ action commands =
     Command mgenInput exec callbacks <-
       Gen.element_ $ filter (\c -> commandGenOK c state0) commands
 
+    -- If we shrink the input, we still want to use the same output. Otherwise
+    -- any actions using this output as part of their input will be dropped. But
+    -- the existing output is still in the context, so `contextNewVar` will
+    -- create a new one. To avoid that, we generate the output before the input.
+    output <- contextNewVar
+
     input <-
       case mgenInput state0 of
         Nothing ->
@@ -555,8 +565,6 @@ action commands =
       pure Nothing
 
     else do
-      output <- contextNewVar
-
       contextUpdate $
         callbackUpdate callbacks state0 input (Var output)
 
