@@ -40,6 +40,7 @@ module Hedgehog.Internal.Tree (
   , filterMaybeT
   , mapMaybeMaybeT
   , filterT
+  , consChild
   , mapMaybeT
   , depth
   , interleave
@@ -316,6 +317,14 @@ mapMaybeT p m =
       Nothing ->
         empty
 
+consChild :: (Monad m) => a -> TreeT m a -> TreeT m a
+consChild a m =
+  TreeT $ do
+    NodeT x xs <- runTreeT m
+    pure $
+      NodeT x $
+        pure a : xs
+
 ------------------------------------------------------------------------
 
 -- | All ways a list can be split
